;; Generated package description from nhexl-mode.el  -*- no-byte-compile: t -*-
(define-package "nhexl-mode" "1.5.0.20220331.114642" "Minor mode to edit files via hex-dump format" '((emacs "24.4")) :commit "45d4611f584a4653c76dbf417ae86d5570a53ea3" :url "https://elpa.gnu.org/packages/nhexl-mode.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca") :keywords '("data"))
